#define G_LOG_DOMAIN "ap-character"

#include "ap-ability.h"
#include "ap-skill.h"
#include "ap-weapon.h"
#include "ap-character.h"

#include <glib.h>
#include <glib-object.h>

typedef struct _APCharacterPrivate
{
    GHashTable *abilities;
    GHashTable *skills;

    GList *binds;

    GHashTable *skill_modifier_ability_bindings;

    GHashTable *binds_skill_modifier;
} APCharacterPrivate;

struct _APCharacter
{
    GObject parent_instance;
};

/**
 * SECTION:ap-character
 * @short_description: Manages communication and registration of adventure-pack
 * objects
 * @title: APCharacter
 * @include: adventure-pack/ap-character.h
 * @see_also: #APAbility, #APSkill, #APAttack, #APItem, #APInventory
 *
 * Manages communication and registration of adventure-pack objects like..
 * * #APAbility
 * * #APSkill
 * * #APAttack
 * * #APItem
 *
 * # Schema # {#schema}
 * |[<!-- language="plain" -->
 * APCharacter
 * ├── abilities
 * │   ╰── ability
 * ├── skills
 * │   ╰── skill
 * ╰── inventory
 *     ╰── item
 * ]|
 *
 * # Purpose {#purpose}
 * ## Abilties and Skills
 *
 * #APCharacter automates the task of linking and updating an #APSkill to an
 * #APAbility. An #APSkill has a property named #APSkill:modifier which relates
 * to the #APAbility:id of an #APAbility. #APSkill<!-- -->s base their
 * #APSkill:score off the #APAbility:modifier of an ability, however,
 * constantly making calls and checks from an #APSkill to an #APAbility is
 * costly. Instead #APCharacter acts as a birdge listing for changes.
 *
 * For example if an abilties #APAbility:score changes and it modifies the
 * ability's #APAbility:modifier then a signal is emitted to #APCharacter.
 * #APCharacter then emits a `::ability-modifier-changed:(ability-id)`
 * signal to any registered #APSkill whos skill #APSKill:modifier matches the
 * ability's #APAbility:id.
 *
 * # Warnings
 * ## Singleton {#singleton-warning}
 * If you plan on use #APCharacter in a singleton manner by using
 * ap_character_get_default() instead of ap_character_new() then you should
 * NEVER unref the object with either g_object_unref() or g_clear_object().
 *
 * In addition #APCharacter is only singleton when used with
 * ap_character_get_default(). Switching between using ap_character_get_default(),
 * ap_character_new(), or creating an %AP_TYPE_CHARACTER with g_object_new() will
 * cause problems. As ap_character_get_default() will use the same global
 * instance and the other two methods will create new #APCharacter objects.
 *
 * # Usage {#usage}
 * ## Complete Example
 * |[<!-- language="C" -->
 * APCharacter *character = ap_character_new ();
 * APAbility *ability;
 * APSkill *skill;
 *
 * ability = ap_ability_new ("dexterity");
 * ap_ability_add_score (ability, 10);
 * ap_character_add_ability (character, ability);
 *
 * skill = ap_skill_new ("stealth");
 * ap_skill_add_modifier ("dexterity");
 * ap_character_add_skill (character, skill);
 *
 * // The "stealth" skill's modifier will now update if the "dexterity" ability
 * // score is changed and causes the "dexterity" ability modifier to change as
 * // the "stealth" skill has added "dexterity" as it's modifier.
 *
 * ability = ap_character_get_ability (character, "dexterity");
 *
 * g_print ("Dexterity Ability Score: %i (%i)",
 *          ap_ability_get_score (ability)
 *          ap_ability_get_modifier (ability));
 *
 * // Returns: Dexterity Ability Score: 10 (0)
 *
 * skill = ap_character_get_skill (character, "stealth");
 *
 * g_print ("Stealth Modifier: %i",
 *          ap_skill_get_score (ability));
 *
 * // Returns: Stealth Modifier: 0
 *
 * ap_ability_set_score (ability, 12);
 *
 * g_print ("Stealth Modifier: %i",
 *          ap_skill_get_score (ability));
 *
 * // Returns: Stealth Modifier: 1
 *
 * // Stealth Modifier updated automatically with signals!
 *
 * g_object_unref (character);
 * ]|
 */

G_DEFINE_TYPE_WITH_PRIVATE (APCharacter, ap_character, G_TYPE_OBJECT)

/* Singleton instance */
APCharacter *ap_character_instance = NULL;

enum
{
    ABILITY_REGISTERED,
    ABILITY_UNREGISTERED,
    SKILL_REGISTERED,
    SKILL_UNREGISTERED,
    WEAPON_REGISTERED,
    WEAPON_UNREGISTERED,
    NUM_SIGNALS
};

static guint signals[NUM_SIGNALS] = { 0, };

// TODO: Evaluate ENUM + new struct approach for GHashTable containing all binds
typedef enum _APCharacterBindType
{
    BIND_ABILITY_SCORE,
    BIND_ABILITY_MODIFIER,
    BIND_SKILL_SCORE,
    BIND_SKILL_MODIFIER
} APCharacterBindType;

typedef struct _APCharacterBind2
{
    gchar *name;
    APCharacterBindType type;

    gpointer source;
    gpointer target;
    GBinding *bind;
} APCharacterBind2;

typedef struct _APCharacterBind
{
    gpointer  source;
    gchar    *source_property;
    gpointer  target;
    gchar    *target_property;
} APCharacterBind;

typedef struct _APCharacterSkill
{
    gpointer  object;
    GBinding *bind_modifier;
} APCharacterSkill;

static void
created_tracked_bind (APCharacter         *self,
                      APCharacterBindType  type,
                      gchar               *name,
                      gpointer             source,
                      gchar               *source_property,
                      gpointer             target,
                      gchar               *target_property,
                      GBindingFlags        flags)
{
    g_assert (AP_IS_CHARACTER (self));

    APCharacterPrivate *priv = ap_character_get_instance_private (self);
    GHashTable *binds;
    GBinding *new_bind;
    GBinding *current_bind;

    switch (type)
    {
        case BIND_SKILL_MODIFIER:
            binds = priv->binds_skill_modifier;
            break;

        default:
            g_warning ("Invalid bind type for created_tracked_bind ().");
            break;
    }

    if (name == NULL)
        g_hash_table_remove (binds, name);

    if (source == NULL)
    {
        g_warning ("Can not create a bind on a source that does not exists");
        return;
    }

    if (target == NULL)
    {
        g_warning ("Can not create a bind on a target that does not exists");
        return;
    }

    current_bind = g_hash_table_lookup (binds, name);

    if (current_bind)
        g_binding_unbind (current_bind);

    new_bind = g_object_bind_property (source, source_property,
                                       target, target_property,
                                       flags);

    g_hash_table_replace (binds, name, new_bind);
}

// TODO: Find a better solution?
static void
on_skill_modifier_changed (GObject    *source,
                           GParamSpec *pspec,
                           gpointer    user_data)
{
    APCharacter *self = AP_CHARACTER (user_data);
    APSkill *skill = AP_SKILL (source);
    gchar *name = ap_skill_get_name (skill);
    gchar *modifier = ap_skill_get_modifier (skill);
    APAbility *ability = ap_character_get_ability (self, modifier);

    if (ability == NULL)
        return;

    created_tracked_bind (self,
                          BIND_SKILL_MODIFIER,
                          name,
                          ability, "modifier",
                          skill, "score",
                          G_BINDING_DEFAULT | G_BINDING_SYNC_CREATE);

    g_free (name);
    g_free (modifier);

    return;












/*
    APCharacter *self = AP_CHARACTER (user_data);
    APCharacterPrivate *priv = ap_character_get_instance_private (self);
    APSkill *skill = AP_SKILL (source);
    gchar *name = ap_skill_get_name (skill);
    gchar *modifier = ap_skill_get_modifier (skill);
    APAbility *ability;

    if (modifier == NULL)
        g_hash_table_remove (priv->skill_modifier_ability_bindings, name);
    else
        ability = ap_character_get_ability (self, modifier);

    if (ability)
    {
        GBinding *new_bind;

        new_bind =
            g_object_bind_property (ability, "modifier",
                                    skill, "score",
                                    G_BINDING_DEFAULT | G_BINDING_SYNC_CREATE);

        g_hash_table_replace (priv->skill_modifier_ability_bindings,
                              name,
                              new_bind);
    }

    g_free (modifier);
*/
}

static gboolean
register_ability (APCharacter *self,
                  APAbility   *ability)
{
    g_assert (AP_IS_CHARACTER (self));
    g_assert (AP_IS_ABILITY (ability));

    APCharacterPrivate *priv = ap_character_get_instance_private (self);
    gchar *name = ap_ability_get_name (ability);

    g_debug ("Attempting to register ability with name \"%s\".", name);

    if (name == NULL)
    {
        g_warning ("Ability not registered, ability must have a name to be registered.");
        return FALSE;
    }

    if (g_hash_table_contains (priv->abilities, name))
    {
        g_warning ("Ability not registered, ability with name \"%s\" already exists.", name);
        return FALSE;
    }

    g_hash_table_insert (priv->abilities, name, AP_ABILITY (g_object_ref (ability)));

    if (!g_hash_table_contains (priv->abilities, name))
    {
        g_warning ("Unable to add ability with name \"%s\" to tracker.", name);
        g_object_unref (ability);
        return FALSE;
    }

    g_signal_emit (self, signals[ABILITY_REGISTERED], 0, ability);

    g_debug ("Ability \"%s\" registered.", name);

    return TRUE;
}

static gboolean
unregister_ability (APCharacter *self,
                    APAbility   *ability)
{
    g_assert (AP_IS_CHARACTER (self));
    g_assert (AP_IS_ABILITY (ability));

    APCharacterPrivate *priv = ap_character_get_instance_private (self);
    gchar *name = ap_ability_get_name (ability);

    g_debug ("Attempting to unregister ability with name \"%s\".", name);

    //TODO: Remove signals

    g_hash_table_remove (priv->abilities, name);

    g_signal_emit (self, signals[ABILITY_UNREGISTERED], 0, ability);

    g_debug ("Ability \"%s\" unregistered.", name);

    g_free (name);

    return TRUE;
}

static gboolean
register_skill (APCharacter *self,
                APSkill   *skill)
{
    g_assert (AP_IS_CHARACTER (self));
    g_assert (AP_IS_SKILL (skill));

    APCharacterPrivate *priv = ap_character_get_instance_private (self);
    gchar *name = ap_skill_get_name (skill);
    gchar *modifier = ap_skill_get_modifier (skill);

    g_debug ("Attempting to register skill with name \"%s\".", name);

    if (name == NULL)
    {
        g_warning ("Skill not registered, skill must have a name to be registered.");
        return FALSE;
    }

    if (g_hash_table_contains (priv->skills, name))
    {
        g_warning ("Skill not registered, skill with name \"%s\" already exists.", name);
        return FALSE;
    }

    g_hash_table_insert (priv->skills, name, AP_SKILL (g_object_ref (skill)));

    if (!g_hash_table_contains (priv->skills, name))
    {
        g_debug ("Unable to add skill with name \"%s\" to tracker.", name);
        g_object_unref (skill);
        return FALSE;
    }

    if (modifier == NULL)
    {
        g_warning ("Skill \"%s\" does not have a modifier, while not required it is recommended to set a modifier before adding it to the character.", name);
    }

    g_signal_connect (skill,
                      "notify::modifier",
                      G_CALLBACK (on_skill_modifier_changed),
                      self);

    if (modifier)
        g_object_notify (G_OBJECT (skill), "modifier");

    g_signal_emit (self, signals[SKILL_REGISTERED], 0, skill);

    g_debug ("Skill \"%s\" registered.", name);

    g_free (modifier);

    return TRUE;
}

static gboolean
unregister_skill (APCharacter *self,
                  APSkill     *skill)
{
    g_assert (AP_IS_CHARACTER (self));
    g_assert (AP_IS_SKILL (skill));

    APCharacterPrivate *priv = ap_character_get_instance_private (self);
    gchar *name = ap_skill_get_name (skill);

    g_debug ("Attempting to unregister skill with name \"%s\".", name);

    //TODO: Remove signals

    g_hash_table_remove (priv->skills, name);

    g_signal_emit (self, signals[SKILL_UNREGISTERED], 0, skill);

    g_debug ("Skill \"%s\" unregistered.", name);

    g_free (name);

    return TRUE;
}

static void
ap_character_dispose (GObject *object)
{
    APCharacter *self = AP_CHARACTER (object);

    ap_character_remove_all (self);

    G_OBJECT_CLASS (ap_character_parent_class)->dispose (object);
}

static void
ap_character_finalize (GObject *object)
{
    APCharacter *self = AP_CHARACTER (object);
    APCharacterPrivate *priv = ap_character_get_instance_private (self);

    if (g_hash_table_size (priv->abilities))
        g_warning ("Not all abilities have been removed.");

    if (g_hash_table_size (priv->skills))
        g_warning ("Not all skills have been removed.");

    g_clear_pointer (&priv->abilities, g_hash_table_unref);
    g_clear_pointer (&priv->skills, g_hash_table_unref);

    G_OBJECT_CLASS (ap_character_parent_class)->finalize (object);
}

static void
ap_character_init (APCharacter *self)
{
    APCharacterPrivate *priv = ap_character_get_instance_private (self);

    priv->abilities = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_object_unref);
    priv->skills = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_object_unref);

    priv->binds_skill_modifier =  g_hash_table_new_full (g_str_hash, g_str_equal, g_free, (GDestroyNotify) g_binding_unbind);
    priv->skill_modifier_ability_bindings = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, (GDestroyNotify) g_binding_unbind);
}

static void
ap_character_class_init (APCharacterClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->dispose = ap_character_dispose;
    object_class->finalize = ap_character_finalize;

    /**
     * APCharacter::ability-registered:
     * @character: an #APCharacter
     * @ability: an #APAbility
     *
     * The ::ability-registered signal is emitted each time an #APAbility is
     * registered.
     */
    signals[ABILITY_REGISTERED] =
        g_signal_new ("ability-registered",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_FIRST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      AP_TYPE_ABILITY);

    /**
     * APCharacter::ability-unregistered:
     * @character: an #APCharacter
     * @ability: an #APAbility
     *
     * The ::ability-unregistered signal is emitted each time an #APAbility is
     * unregistered.
     */
    signals[ABILITY_UNREGISTERED] =
        g_signal_new ("ability-unregistered",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_FIRST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      AP_TYPE_ABILITY);

    /**
     * APCharacter::skill-registered:
     * @character: an #APCharacter
     * @skill: an #APSkill
     *
     * The ::ability-registered signal is emitted each time an #APSkill is
     * registered.
     */
    signals[SKILL_REGISTERED] =
        g_signal_new ("skill-registered",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_FIRST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      AP_TYPE_SKILL);

    /**
     * APCharacter::skill-unregistered:
     * @character: an #APCharacter
     * @skill: an #APSkill
     *
     * The ::ability-unregistered signal is emitted each time an #APSkill is
     * unregistered.
     */
    signals[SKILL_UNREGISTERED] =
        g_signal_new ("skill-unregistered",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_FIRST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      AP_TYPE_SKILL);
}

/*
 * *****************************************************************************
 * API
 * *****************************************************************************
 */
/**
 * ap_character_new:
 *
 * Creates a new #APCharacter which manages the connections between
 * #APAbility<!-- -->s, #APSkill<!-- -->s, #APAttack<!-- -->s,
 * #APItem<!-- -->s, etc... #APCharacter registers and unregisters the various
 * signals for communication between them.
 *
 * # Usage
 * |[<!-- language="C" -->
 * APCharacter *character;
 *
 * character = ap_character_new ();
 *
 * // Do some stuff...
 *
 * g_object_unref (character);
 * ]|
 *
 * Returns: (transfer full): an #APCharacter instance
 */
APCharacter *
ap_character_new (void)
{
    return g_object_new (AP_TYPE_CHARACTER,
                         NULL);
}

/**
 * ap_character_get_default:
 *
 * Returns the default singleton #APCharacter instance, if a global #APCharacter
 * singleton instance does not exist then one is created.
 *
 * # Warning
 * If you use ap_character_get_default() at all then you are agreeing to use
 * #APCharacter in singleton mode. See [here][singleton-warning] for more
 * information.
 *
 * # Usage
 * |[<!-- language="C" -->
 * APCharacter *character;
 *
 * character = ap_character_get_default ();
 * ]|
 *
 * Returns: (transfer none): the singleton #APCharacter instance
 */
APCharacter *
ap_character_get_default (void)
{
    if (!ap_character_instance)
        ap_character_instance = ap_character_new ();

    return ap_character_instance;
}

/**
 * ap_character_add_ability:
 * @self: an #APCharacter
 * @ability: an #APAbility
 *
 * Adds an #APAbility to the #APCharacter.
 *
 * Returns: (gboolean): %TRUE if @ability has been added
 */
gboolean
ap_character_add_ability (APCharacter *self,
                          APAbility   *ability)
{
    g_assert (AP_IS_CHARACTER (self));
    g_assert (AP_IS_ABILITY (ability));

    return register_ability (self, ability);
}

/**
 * ap_character_get_ability:
 * @self: an #APCharacter
 * @name: the name of an #APAbility
 *
 * Finds and returns an ability with @name, can return %NULL if no ability with
 * @name is found.
 *
 * Returns: (transfer none) (nullable): an #APAbility
 */
APAbility *
ap_character_get_ability (APCharacter *self,
                          gchar       *name)
{
    APCharacterPrivate *priv;

    g_assert (AP_IS_CHARACTER (self));

    priv = ap_character_get_instance_private (self);

    return g_hash_table_lookup (priv->abilities, name);
}

/**
 * ap_character_remove_ability:
 * @self: an #APCharacter
 * @name: the name of an #APAbility
 *
 * Removes an #APAbility from the #APCharacter.
 *
 * Returns: (gboolean): %TRUE if the #APAbility with @name was found and removed
 *          or if the #APAbility is not currently tracked by #APCharacter
 */
gboolean
ap_character_remove_ability (APCharacter *self,
                             gchar       *name)
{
    APAbility *ability;

    g_assert (AP_IS_CHARACTER (self));

    ability = ap_character_get_ability (self, name);

    if (ability == NULL)
        return TRUE;

    return unregister_ability (self, ability);
}

/**
 * ap_character_add_skill:
 * @self: an #APCharacter
 * @skill: an #APSkill
 *
 * Adds an #APSkill to the #APCharacter.
 *
 * Returns: (gboolean): %TRUE if @skill has been added
 */
gboolean
ap_character_add_skill (APCharacter *self,
                        APSkill     *skill)
{
    g_assert (AP_IS_CHARACTER (self));
    g_assert (AP_IS_SKILL (skill));

    return register_skill (self, skill);
}

/**
 * ap_character_get_skill:
 * @self: an #APCharacter
 * @name: the name of an #APSkill
 *
 * Finds and returns a skill with @name, can return %NULL if no skill with
 * @name is found.
 *
 * Returns: (transfer none) (nullable): an #APSkill
 */
APSkill *
ap_character_get_skill (APCharacter *self,
                        gchar       *name)
{
    APCharacterPrivate *priv;

    g_assert (AP_IS_CHARACTER (self));

    priv = ap_character_get_instance_private (self);

    return g_hash_table_lookup (priv->skills, name);
}

/**
 * ap_character_remove_skill:
 * @self: an #APCharacter
 * @name: the name of an #APSkill
 *
 * Removes an #APSkill from the #APCharacter.
 *
 * Returns: (gboolean): %TRUE if the #APSkill with @name was found and removed
 *          or if the #APSkill is not currently tracked by #APCharacter
 */
gboolean
ap_character_remove_skill (APCharacter *self,
                           gchar       *name)
{
    APSkill *skill;

    g_assert (AP_IS_CHARACTER (self));

    skill = ap_character_get_skill (self, name);

    if (skill == NULL)
        return TRUE;

    return unregister_skill (self, skill);
}

/**
 * ap_character_get_abilities:
 * @self: an #APCharacter
 *
 * Returns a #GList containing every #APAbility registered.
 *
 * Returns: (transfer container): a #GList containing all the abilities. The
 *          content of the list is owned by #APCharacter and should not be
 *          modified or freed. Use g_list_free() when done using the list.
 */
GList *
ap_character_get_abilities (APCharacter *self)
{
    APCharacterPrivate *priv;

    g_assert (AP_IS_CHARACTER (self));

    priv = ap_character_get_instance_private (self);

    return g_hash_table_get_values (priv->abilities);
}

/**
 * ap_character_remove_abilities:
 * @self: an #APCharacter
 *
 * Removes all abilities from @self.
 */
void
ap_character_remove_abilities (APCharacter *self)
{
    GList *abilities;
    GList *list;

    g_assert (AP_IS_CHARACTER (self));

    abilities = ap_character_get_abilities (self);

    for (list = abilities; list != NULL; list = list->next)
    {
        g_assert (AP_IS_ABILITY (list->data));

        unregister_ability (self, list->data);
    }

    g_list_free (abilities);
}

/**
 * ap_character_get_skills:
 * @self: an #APCharacter
 *
 * Returns a #GList containing every #APSkill registered.
 *
 * Returns: (transfer container): a #GList containing all the skills. The
 *          content of the list is owned by #APCharacter and should not be
 *          modified or freed. Use g_list_free() when done using the list.
 */
GList *
ap_character_get_skills (APCharacter *self)
{
    APCharacterPrivate *priv;

    g_assert (AP_IS_CHARACTER (self));

    priv = ap_character_get_instance_private (self);

    return g_hash_table_get_values (priv->skills);
}

/**
 * ap_character_remove_skills:
 * @self: an #APCharacter
 *
 * Removes all skills from @self.
 */
void
ap_character_remove_skills (APCharacter *self)
{
    GList *skills;
    GList *list;

    g_assert (AP_IS_CHARACTER (self));

    skills = ap_character_get_skills (self);

    for (list = skills; list != NULL; list = list->next)
    {
        g_assert (AP_IS_SKILL (list->data));

        unregister_skill (self, list->data);
    }

    g_list_free (skills);
}

/**
 * ap_character_remove_all:
 * @self: an #APCharacter
 *
 * Removes all #APAbilitys, #APSkills, etc.. from the #APCharacter.
 */
void
ap_character_remove_all (APCharacter *self)
{
    g_assert (AP_IS_CHARACTER (self));

    ap_character_remove_abilities (self);
    ap_character_remove_skills (self);
}
