#define G_LOG_DOMAIN "ap-weapon"

#include "ap-weapon.h"

#include <glib.h>
#include <glib-object.h>

typedef struct _APWeaponPrivate
{
    gchar    *name;
    gchar    *damage;          // Dice type, ex: 1d4, 1d6, 1d8, 1d10, 2d6, etc..
    gchar    *type;            // bludgeoning, slashing, piercing, etc..

    gchar    *proficiency;     // simple/martial

    gboolean  light;           // Is a light weapon
    gboolean  heavy;           // Is a heavy weapon
    gboolean  finesse;         // Is a finesse weapon
    gboolean  ranged;          // Is a ranged weapon
    gboolean  reach;           // Adds 5 feet reach
    gboolean  loaded;          // Is a weapon that needs to be loaded each use
    gboolean  thrown;          // Is a throwable weapon
    gboolean  two_handed;      // Is a two-handed weapon
    gboolean  versatile;       // Is a versatile weapon and can be used with one or two hands
    gboolean  silvered;        // Is a silver coated weapon
    gboolean  improvised;      // Is an improvised weapon

    gint      range;
    gint      range_long;

    gint      ammunition;
    gchar    *ammunition_name;

    gchar    *versatile_damge;

    gboolean  special;
    char     *special_description;
} APWeaponPrivate;

struct _APWeapon
{
    GObject parent_instance;
};

/**
 * SECTION:ap-ability
 * @short_description: Manages a single ability of a character
 * @title: APAbility
 * @include: adventure-pack/ap-ability.h
 *
 *
 *
 * # Usage {#usage}
 * ## Basic
 * |[<!-- language="C" -->
 * APWeapon *weapon;
 *
 * weapon = g_object_new (AP_TYPE_WEAPON,
 *                        "name", "Longsword",
 *                        "damage", "1d8",
 *                        "type", "Slashing"
 *                        NULL);
 *
 * g_print ("%s: %i (%i)",
 *          ap_weapon_get_name (weapon),
 *          ap_weapon_get_damage (weapon),
 *          ap_weapon_get_critical (weapon));
 *
 * // Returns: Slash: 10 (0)
 * ]|
 *
 * ## Add properties
 * |[<!-- language="C" -->
 * APWeapon *weapon;
 *
 * weapon = ap_weapon_new ("Slash");
 * ap_weapon_add_damage (weapon, 10);
 * ]|
 */

G_DEFINE_TYPE_WITH_PRIVATE (APWeapon, ap_weapon, G_TYPE_OBJECT)

enum
{
    PROP_0,
    PROP_NAME,
    PROP_LABEL,
    PROP_TYPE,
    PROP_RANGE,
    PROP_DAMAGE,
    PROP_CRITICAL,
    PROP_AMMO,
    PROP_BONUS,
    NUM_PROPERTIES
};

static GParamSpec *properties[NUM_PROPERTIES] = { NULL, };

static void
ap_weapon_set_property (GObject      *object,
                           guint         property_id,
                           const GValue *value,
                           GParamSpec   *pspec)
{
    APWeapon *self = AP_WEAPON (object);
    APWeaponPrivate *priv = ap_weapon_get_instance_private (self);

    switch (property_id)
    {
        case PROP_NAME:
            priv->name = g_value_dup_string (value);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

static void
ap_weapon_get_property (GObject      *object,
                           guint         property_id,
                           GValue       *value,
                           GParamSpec   *pspec)
{
    APWeapon *self = AP_WEAPON (object);
    APWeaponPrivate *priv = ap_weapon_get_instance_private (self);

    switch (property_id)
    {
        case PROP_NAME:
            g_value_set_string (value, priv->name);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

static void
ap_weapon_finalize (GObject *object)
{
   // APWeaponPrivate *priv = ap_weapon_get_instance_private (AP_WEAPON (object));

    G_OBJECT_CLASS (ap_weapon_parent_class)->finalize (object);
}

static void
ap_weapon_dispose (GObject *object)
{
    //APWeaponPrivate *priv = ap_weapon_get_instance_private (AP_WEAPON (object));

    G_OBJECT_CLASS (ap_weapon_parent_class)->finalize (object);
}

static void
ap_weapon_init (APWeapon *self)
{
    //APWeaponPrivate *priv = ap_weapon_get_instance_private (self);
}

static void
ap_weapon_class_init (APWeaponClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->set_property = ap_weapon_set_property;
    object_class->get_property = ap_weapon_get_property;
    object_class->finalize = ap_weapon_finalize;
    object_class->dispose = ap_weapon_dispose;

	/**
	 * APWeapon:name:
	 *
	 * Name of the weapon.
     *
     * # Example
     * * `Dexterity`
	 */
    properties[PROP_NAME] =
        g_param_spec_string ("name",
                             "name",
                             "Name of the weapon.",
                             NULL,
                             G_PARAM_READWRITE);

    g_object_class_install_properties (object_class,
                                       NUM_PROPERTIES,
                                       properties);
}

/*
 * *****************************************************************************
 * API
 * *****************************************************************************
 */
/**
 * ap_weapon_new:
 *
 * Creates a new #APWeapon.
 *
 * Returns: (transfer full): an #APWeapon instance
 */
APWeapon *
ap_weapon_new (void)
{
    return g_object_new (AP_TYPE_WEAPON,
                         NULL);
}
/**
 * ap_weapon_new_with_name:
 * @name: the name of the weapon
 *
 * Creates a new #APWeapon.
 *
 * Returns: (transfer full): an #APWeapon instance
 */
APWeapon *
ap_weapon_new_with_name (gchar *name)
{
    return g_object_new (AP_TYPE_WEAPON,
                         "name", name,
                         NULL);
}

/**
 * ap_weapon_new_simple:
 * @name: the name of the weapon
 * @damage: the dice used for role, i.e. 1d4, 1d6, 1d8, 2d4, etc..
 * @type: the damage type, i.e. slashing, bludgeoning, etc..
 *
 * Creates a new #APWeapon.
 *
 * Returns: (transfer full): an #APWeapon instance
 */
APWeapon *
ap_weapon_new_simple (gchar *name,
                      gchar *damage,
                      gchar *type)
{
    return g_object_new (AP_TYPE_WEAPON,
                         "name", name,
                         "damage", damage,
                         "type", type,
                         NULL);
}
